from torchline.utils import Registry

META_ARCH_REGISTRY = Registry("META_ARCH")
META_ARCH_REGISTRY.__doc__ = """
Registry for meta-architectures, i.e. the whole model.

The registered object will be called with `obj(cfg)`
and expected to return a `nn.Module` object.
"""


def build_model(cfg):
    """
    Built the whole model, defined by `cfg.MODEL.META_ARCH`.
    """
    meta_arch = cfg.MODEL.META_ARCH
    return META_ARCH_REGISTRY.get(meta_arch)(cfg)
    


from torchline.utils import Registry

LOSS_FN_REGISTRY = Registry("LOSS_FN")
LOSS_FN_REGISTRY.__doc__ = """
Registry for loss function, e.g. cross entropy loss.

The registered object will be called with `obj(cfg)`
"""


def build_loss_fn(cfg):
    """
    Built the loss function, defined by `cfg.LOSS.NAME`.
    """
    name = cfg.LOSS.NAME
    return LOSS_FN_REGISTRY.get(name)(cfg)



from .build import build_data, DATASET_REGISTRY
from .skin100 import Skin100Dataset
from .skin10 import Skin10Dataset
from .common_datasets import MNIST, CIFAR10
from .transforms import DefaultTransforms, build_transforms, build_label_transforms, TRANSFORMS_REGISTRY, LABEL_TRANSFORMS_REGISTRY
from .autoaugment import ImageNetPolicy, CIFAR10Policy, SVHNPolicy, SubPolicy
from .sampler import build_sampler, SAMPLER_REGISTRY


"""
Example template for defining a system
"""
import logging
import os
from argparse import ArgumentParser
from collections import OrderedDict

import pytorch_lightning as pl
import torch
import torch.nn as nn
import torch.nn.functional as F
import torchvision.transforms as transforms
from pytorch_lightning import LightningModule
from torch import optim
from torch.utils.data import DataLoader
from torch.utils.data.distributed import DistributedSampler
from torchvision.datasets import MNIST

from torchline.data import build_data, build_sampler
from torchline.losses import build_loss_fn
from torchline.models import build_model
from .build import MODULE_TEMPLATE_REGISTRY

@MODULE_TEMPLATE_REGISTRY.register()
class LightningTemplateModel(LightningModule):
    """
    Sample model to show how to define a template
    """

    def __init__(self, cfg):
        """
        Pass in parsed HyperOptArgumentParser to the model
        :param cfg:
        """
        # init superclass
        super(LightningTemplateModel, self).__init__()
        self.cfg = cfg

        self.batch_size = self.cfg.DATASET.BATCH_SIZE

        # if you specify an example input, the summary will show input/output for each layer
        h, w = self.cfg.INPUT.SIZE
        self.example_input_array = torch.rand(1, 3, h, w)

        # build model
        self.model = self.build_model(cfg)
        self.loss_fn = self.build_loss_fn(cfg)

    # ---------------------
    # MODEL SETUP
    # ---------------------
    @classmethod
    def build_model(cls, cfg):
        """
        Layout model
        :return:
        """
        return build_model(cfg)

    @classmethod
    def build_loss_fn(cls, cfg):
        """
        Layout loss_fn
        :return:
        """
        return build_loss_fn(cfg)

    @classmethod
    def build_data(cls, cfg, is_train):
        """
        Layout training dataset
        :return:
        """
        cfg.defrost()
        cfg.DATASET.IS_TRAIN = is_train
        cfg.freeze()
        return build_data(cfg)

    @classmethod
    def build_sampler(cls, cfg, is_train):
        """
        Layout training dataset
        :return:
        """
        cfg.defrost()
        cfg.DATASET.IS_TRAIN = is_train
        cfg.freeze()
        return build_sampler(cfg)

    # ---------------------
    # TRAINING
    # ---------------------
    
    def forward(self, x):
        """
        No special modification required for lightning, define as you normally would
        :param x:
        :return:
        """

        logits = self.model(x)
        return logits

    def loss(self, predictions, gt_labels):
        return self.loss_fn(predictions, gt_labels)

    def training_step(self, batch, batch_idx):
        """
        Lightning calls this inside the training loop
        :param batch:
        :return:
        """
        # forward pass
        inputs, gt_labels = batch
        predictions = self.forward(inputs)

        # calculate loss
        loss_val = self.loss(predictions, gt_labels)

        # acc
        predict_labels = torch.argmax(predictions, dim=1)
        train_acc = torch.sum(gt_labels== predict_labels).item() / (len(gt_labels) * 1.0)
        train_acc = torch.tensor(train_acc)

        # in DP mode (default) make sure if result is scalar, there's another dim in the beginning
        if self.trainer.use_dp or self.trainer.use_ddp2:
            loss_val = loss_val.unsqueeze(0)

        tqdm_dict = {'train_loss': loss_val, 'train_acc': train_acc}
        output = OrderedDict({
            'loss': loss_val,
            'progress_bar': tqdm_dict,
            'log': tqdm_dict
        })


        # can also return just a scalar instead of a dict (return loss_val)
        return output

    def validation_step(self, batch, batch_idx):
        """
        Lightning calls this inside the validation loop
        :param batch:
        :return:
        """
        inputs, gt_labels = batch
        predictions = self.forward(inputs)

        loss_val = self.loss(predictions, gt_labels)

        # acc
        predict_labels = torch.argmax(predictions, dim=1)
        val_acc = torch.sum(gt_labels== predict_labels).item() / (len(gt_labels) * 1.0)
        val_acc = torch.tensor(val_acc)

        if self.on_gpu:
            val_acc = val_acc.cuda(loss_val.device.index)

        # in DP mode (default) make sure if result is scalar, there's another dim in the beginning
        if self.trainer.use_dp or self.trainer.use_ddp2:
            loss_val = loss_val.unsqueeze(0)
            val_acc = val_acc.unsqueeze(0)

        output = OrderedDict({
            'val_loss': loss_val,
            'val_acc': val_acc,
        })

        # can also return just a scalar instead of a dict (return loss_val)
        return output

    def validation_end(self, outputs):
        """
        Called at the end of validation to aggregate outputs
        :param outputs: list of individual outputs of each validation step
        :return:
        """
        # if returned a scalar from validation_step, outputs is a list of tensor scalars
        # we return just the average in this case (if we want)
        # return torch.stack(outputs).mean()

        val_loss_mean = 0
        val_acc_mean = 0
        for output in outputs:
            val_loss = output['val_loss']

            # reduce manually when using dp
            if self.trainer.use_dp or self.trainer.use_ddp2:
                val_loss = torch.mean(val_loss)
            val_loss_mean += val_loss

            # reduce manually when using dp
            val_acc = output['val_acc']
            if self.trainer.use_dp or self.trainer.use_ddp2:
                val_acc = torch.mean(val_acc)

            val_acc_mean += val_acc

        val_loss_mean /= len(outputs)
        val_acc_mean /= len(outputs)
        tqdm_dict = {'val_loss': val_loss_mean, 'val_acc': val_acc_mean}
        result = {'progress_bar': tqdm_dict, 'log': tqdm_dict, 'val_loss': val_loss_mean}
        return result

    # ---------------------
    # TRAINING SETUP
    # ---------------------
    def generate_optimizer(self):
        '''
        return torch.optim.Optimizer
        '''
        optim_name = self.cfg.OPTIM.NAME
        momentum = self.cfg.OPTIM.MOMENTUM
        weight_decay = self.cfg.OPTIM.WEIGHT_DECAY
        lr = self.cfg.OPTIM.BASE_LR
        if optim_name.lower() == 'sgd':
            return torch.optim.SGD(self.parameters(), lr=lr, momentum=momentum, weight_decay=weight_decay)
        elif optim_name.lower() == 'adadelta':
            return torch.optim.Adagrad(self.parameters(), lr=lr, weight_decay=weight_decay)
        elif optim_name.lower() == 'adam': 
            return torch.optim.Adam(self.parameters(), lr=lr, weight_decay=weight_decay)
        elif optim_name.lower() == 'rmsprop':
            return torch.optim.RMSprop(self.parameters(), lr=lr, momentum=momentum, weight_decay=weight_decay)
        else:
            print(f"{optim_name} not implemented")
            raise NotImplementedError

    def generate_scheduler(self, optimizer):
        '''
        return torch.optim.lr_scheduler
        '''
        scheduler_name = self.cfg.OPTIM.SCHEDULER.NAME
        if scheduler_name.lower() == 'CosineAnnealingLR'.lower():
            return optim.lr_scheduler.CosineAnnealingLR(optimizer, T_max=10)
        elif scheduler_name.lower() == 'StepLR'.lower():
            step_size = self.cfg.OPTIM.SCHEDULER.STEP_SIZE
            gamma = self.cfg.OPTIM.SCHEDULER.GAMMA
            return optim.lr_scheduler.StepLR(optimizer, step_size=step_size, gamma=gamma)
        elif scheduler_name.lower() == 'MultiStepLR'.lower():
            milestones = self.cfg.OPTIM.SCHEDULER.MILESTONES
            gamma = self.cfg.OPTIM.SCHEDULER.GAMMA
            return optim.lr_scheduler.MultiStepLR(optimizer, milestones=milestones, gamma=gamma)
        elif scheduler_name.lower() == 'ReduceLROnPlateau'.lower():
            mode = self.cfg.OPTIM.SCHEDULER.MODE
            patience = self.cfg.OPTIM.SCHEDULER.PATIENCE
            verbose = self.cfg.OPTIM.SCHEDULER.VERBOSE
            factor = self.cfg.OPTIM.SCHEDULER.GAMMA
            return optim.lr_scheduler.ReduceLROnPlateau(optimizer, mode=mode, patience=patience, 
                                                        verbose=verbose, factor=factor)
        else:
            print(f"{scheduler_name} not implemented")
            raise NotImplementedError

    def configure_optimizers(self):
        """
        return whatever optimizers we want here
        :return: list of optimizers
        """
        optimizer = self.generate_optimizer()
        scheduler = self.generate_scheduler(optimizer)
        return [optimizer], [scheduler]

    def __dataloader(self, is_train):
        # init data generators
        dataset = self.build_data(self.cfg, is_train)

        # when using multi-node (ddp) we need to add the  datasampler
        train_sampler = self.build_sampler(self.cfg, is_train)
        batch_size = self.batch_size

        if self.use_ddp:
            train_sampler = DistributedSampler(dataset)

        should_shuffle = train_sampler is None
        loader = DataLoader(
            dataset=dataset,
            batch_size=batch_size,
            shuffle=should_shuffle,
            sampler=train_sampler,
            num_workers=self.cfg.DATALOADER.NUM_WORKERS
        )

        return loader

    @pl.data_loader
    def train_dataloader(self):
        logging.info('training data loader called')
        return self.__dataloader(is_train=True)

    @pl.data_loader
    def val_dataloader(self):
        logging.info('val data loader called')
        return self.__dataloader(is_train=False)

    @pl.data_loader
    def test_dataloader(self):
        logging.info('test data loader called')
        return self.__dataloader(is_train=False)



#coding=utf-8
import torch
import numpy as np
import math
import random

## todo
# def softlabel(label):
#     return label

class Cutout(object):
    """Randomly mask out one or more patches from an image.
    Args:
        n_holes (int): Number of patches to cut out of each image.
        length (int): The length (in pixels) of each square patch.
    """
    def __init__(self, n_holes=1, length=16):
        self.n_holes = n_holes
        self.length = length

    def __call__(self, img):
        """
        Args:
            img (Tensor): Tensor image of size (C, H, W).
        Returns:
            Tensor: Image with n_holes of dimension length x length cut out of it.
        """
        h = img.size(1)
        w = img.size(2)

        mask = np.ones((h, w), np.float32)

        for n in range(self.n_holes):
            y = np.random.randint(h)
            x = np.random.randint(w)

            y1 = np.clip(y - self.length // 2, 0, h)
            y2 = np.clip(y + self.length // 2, 0, h)
            x1 = np.clip(x - self.length // 2, 0, w)
            x2 = np.clip(x + self.length // 2, 0, w)

            mask[y1: y2, x1: x2] = 0.

        mask = torch.from_numpy(mask)
        mask = mask.expand_as(img)
        img = img * mask

        return img

class RandomErasing(object):
    '''
    Class that performs Random Erasing in Random Erasing Data Augmentation by Zhong et al. 
    -------------------------------------------------------------------------------------
    probability: The probability that the operation will be performed.
    sl: min erasing area
    sh: max erasing area
    r1: min aspect ratio
    mean: erasing value
    -------------------------------------------------------------------------------------
    '''
    def __init__(self, probability = 0.5, sl = 0.02, sh = 0.4, r1 = 0.3, mean=[0.4914, 0.4822, 0.4465]):
        self.probability = probability
        self.mean = mean
        self.sl = sl
        self.sh = sh
        self.r1 = r1
       
    def __call__(self, img):

        if random.uniform(0, 1) > self.probability:
            return img

        for _ in range(100):
            area = img.size()[1] * img.size()[2]
       
            target_area = random.uniform(self.sl, self.sh) * area
            aspect_ratio = random.uniform(self.r1, 1/self.r1)

            h = int(round(math.sqrt(target_area * aspect_ratio)))
            w = int(round(math.sqrt(target_area / aspect_ratio)))

            if w < img.size()[2] and h < img.size()[1]:
                x1 = random.randint(0, img.size()[1] - h)
                y1 = random.randint(0, img.size()[2] - w)
                if img.size()[0] == 3:
                    img[0, x1:x1+h, y1:y1+w] = self.mean[0]
                    img[1, x1:x1+h, y1:y1+w] = self.mean[1]
                    img[2, x1:x1+h, y1:y1+w] = self.mean[2]
                else:
                    img[0, x1:x1+h, y1:y1+w] = self.mean[0]
                return img

        return img


import torch
import torch.nn.functional as F
from .build import LOSS_FN_REGISTRY

__all__ = [
    'cross_entropy'
]

@LOSS_FN_REGISTRY.register()
class CrossEntropy:
    def __init__(self, cfg):
        self.loss_fn = F.cross_entropy
    
    def __call__(self, predictions, gt_labels):
        return self.loss_fn(predictions, gt_labels, reduction="mean")



from torchvision.datasets import MNIST as _MNIST
from torchvision.datasets import CIFAR10 as _CIFAR10
from .build import DATASET_REGISTRY
from .transforms import build_transforms


__all__ = [
    'MNIST',
    'CIFAR10',
]

@DATASET_REGISTRY.register()
def MNIST(cfg):
    root = cfg.DATASET.DIR
    is_train = cfg.DATASET.IS_TRAIN
    transform = build_transforms(cfg)
    return _MNIST(root=root, train=is_train, transform=transform.transform, download=True)

@DATASET_REGISTRY.register()
def CIFAR10(cfg):
    root = cfg.DATASET.DIR
    is_train = cfg.DATASET.IS_TRAIN
    transform = build_transforms(cfg)
    return _CIFAR10(root=root, train=is_train, transform=transform.transform, download=True)


from yacs.config import CfgNode as _CfgNode

class CfgNode(_CfgNode):
    pass

global_cfg = CfgNode()

def get_cfg():
    '''
    Get a copy of the default config.

    Returns:
        a CfgNode instance.
    '''
    from .default import _C
    return _C.clone()



from torchline.utils import Registry, Logger

MODULE_TEMPLATE_REGISTRY = Registry('MODULE_TEMPLATE')
MODULE_TEMPLATE_REGISTRY.__doc__ = """
Registry for module template, e.g. LightningTemplateModel.

The registered object will be called with `obj(cfg)`
"""

def build_module_template(cfg):
    """
    Built the module template, defined by `cfg.MODULE_TEMPLATE.NAME`.
    """
    name = cfg.MODULE_TEMPLATE.NAME
    return MODULE_TEMPLATE_REGISTRY.get(name)(cfg)

    


from .config import CfgNode as CN

_C = CN()
_C.VERSION = 1


# ---------------------------------------------------------------------------- #
# input
# ---------------------------------------------------------------------------- #
_C.INPUT = CN()
_C.INPUT.SIZE = (224, 224)

# ---------------------------------------------------------------------------- #
# dataset
# ---------------------------------------------------------------------------- #
_C.DATASET = CN()
_C.DATASET.NAME = 'cifar10'
_C.DATASET.BATCH_SIZE = 16
_C.DATASET.DIR = './datasets/skin100_dataset/train'
_C.DATASET.TRAIN_LIST = './datasets/train_skin10.txt'
_C.DATASET.VALID_LIST = './datasets/valid_skin10.txt'
_C.DATASET.TEST_LIST = './datasets/test_skin10.txt'
_C.DATASET.IS_TRAIN = False # specify to load training or testing set


# ---------------------------------------------------------------------------- #
# transforms
# ---------------------------------------------------------------------------- #

_C.TRANSFORMS = CN() # image transforms
_C.TRANSFORMS.NAME = 'DefaultTransforms'


## transforms for tensor
_C.TRANSFORMS.TENSOR = CN()
# for skin100
_C.TRANSFORMS.TENSOR.NORMALIZATION = CN()
_C.TRANSFORMS.TENSOR.NORMALIZATION.mean = [0.6075, 0.4564, 0.4182] 
_C.TRANSFORMS.TENSOR.NORMALIZATION.std = [0.2158, 0.1871, 0.1826]
# _C.TRANSFORMS.TENSOR.NORMALIZATION = {
#             'mean':[0.6054, 0.4433, 0.4084], 
#             'std': [0.2125, 0.1816, 0.1786]  # for skin10
_C.TRANSFORMS.TENSOR.RANDOM_ERASING = CN()
_C.TRANSFORMS.TENSOR.RANDOM_ERASING.enable = 0
_C.TRANSFORMS.TENSOR.RANDOM_ERASING.p = 0.5
_C.TRANSFORMS.TENSOR.RANDOM_ERASING.scale = (0.02, 0.3) # range of proportion of erased area against input image.
_C.TRANSFORMS.TENSOR.RANDOM_ERASING.ratio = (0.3, 3.3), # range of aspect ratio of erased area.


## transforms for PIL image
_C.TRANSFORMS.IMG = CN()

### modify the image size, only use one operation
# random_resized_crop
_C.TRANSFORMS.IMG.RANDOM_RESIZED_CROP = CN()
_C.TRANSFORMS.IMG.RANDOM_RESIZED_CROP.enable = 0
_C.TRANSFORMS.IMG.RANDOM_RESIZED_CROP.scale = (0.5, 1.0)
_C.TRANSFORMS.IMG.RANDOM_RESIZED_CROP.ratio = (3/4, 4/3)

# resize
_C.TRANSFORMS.IMG.RESIZE =  CN()
_C.TRANSFORMS.IMG.RESIZE.enable = 1

# random_crop
_C.TRANSFORMS.IMG.RANDOM_CROP = CN()
_C.TRANSFORMS.IMG.RANDOM_CROP.enable = 1

# center_crop
_C.TRANSFORMS.IMG.CENTER_CROP = CN()
_C.TRANSFORMS.IMG.CENTER_CROP.enable = 0

### without modifying the image size
_C.TRANSFORMS.IMG.AUG_IMAGENET = False
_C.TRANSFORMS.IMG.AUG_CIFAR = False

# color_jitter
_C.TRANSFORMS.IMG.COLOR_JITTER = CN()
_C.TRANSFORMS.IMG.COLOR_JITTER.brightness = 0.
_C.TRANSFORMS.IMG.COLOR_JITTER.contrast = 0.
_C.TRANSFORMS.IMG.COLOR_JITTER.saturation = 0.
_C.TRANSFORMS.IMG.COLOR_JITTER.hue = 0.

# horizontal_flip
_C.TRANSFORMS.IMG.RANDOM_HORIZONTAL_FLIP = CN()
_C.TRANSFORMS.IMG.RANDOM_HORIZONTAL_FLIP.enable = 1
_C.TRANSFORMS.IMG.RANDOM_HORIZONTAL_FLIP.p = 1

# vertical_flip
_C.TRANSFORMS.IMG.RANDOM_VERTICAL_FLIP = CN()
_C.TRANSFORMS.IMG.RANDOM_VERTICAL_FLIP.enable = 1
_C.TRANSFORMS.IMG.RANDOM_VERTICAL_FLIP.p = 1

# random_rotation
_C.TRANSFORMS.IMG.RANDOM_ROTATION = CN()
_C.TRANSFORMS.IMG.RANDOM_ROTATION.enable = 1
_C.TRANSFORMS.IMG.RANDOM_ROTATION.degrees = 15



_C.LABEL_TRANSFORMS = CN() # label transforms
_C.LABEL_TRANSFORMS.NAME = 'default'


# ---------------------------------------------------------------------------- #
# dataloader
# ---------------------------------------------------------------------------- #
_C.DATALOADER = CN()
_C.DATALOADER.NUM_WORKERS = 4
_C.DATALOADER.SAMPLER_TRAIN = "default"
_C.DATALOADER.SAMPLER_TEST = "default"


# ---------------------------------------------------------------------------- #
# model
# ---------------------------------------------------------------------------- #
_C.MODEL = CN()
_C.MODEL.META_ARCH = 'Resnet50'
_C.MODEL.WEIGHTS = ""
_C.MODEL.CLASSES = 10
_C.MODEL.PRETRAINED = True
_C.MODEL.FEATURES = ['f4', ]
_C.MODEL.FEATURES_FUSION = 'sum'


# ---------------------------------------------------------------------------- #
# optimizer
# ---------------------------------------------------------------------------- #
_C.OPTIM = CN()
_C.OPTIM.NAME = 'adam'
_C.OPTIM.MOMENTUM = 0.9
_C.OPTIM.BASE_LR = 0.001
_C.OPTIM.WEIGHT_DECAY = 0.0001

# scheduler
_C.OPTIM.SCHEDULER = CN()
_C.OPTIM.SCHEDULER.NAME = 'CosineAnnealingLR'
_C.OPTIM.SCHEDULER.GAMMA = 0.1 # decay factor

# for CosineAnnealingLR
_C.OPTIM.SCHEDULER.T_MAX = 10 

# for ReduceLROnPlateau
_C.OPTIM.SCHEDULER.MODE = 'min' # min for loss, max for acc
_C.OPTIM.SCHEDULER.PATIENCE = 10
_C.OPTIM.SCHEDULER.VERBOSE = True # print log once update lr

# for StepLR
_C.OPTIM.SCHEDULER.STEP_SIZE = 10

# for MultiStepLR
_C.OPTIM.SCHEDULER.MILESTONES = [10, 15, 25, 35]

# ---------------------------------------------------------------------------- #
# loss
# ---------------------------------------------------------------------------- #
_C.LOSS = CN()
_C.LOSS.NAME = 'CrossEntropy'

# ---------------------------------------------------------------------------- #
# hooks
# ---------------------------------------------------------------------------- #
_C.HOOKS = CN()

## EarlyStopping
_C.HOOKS.EARLY_STOPPING = CN()
_C.HOOKS.EARLY_STOPPING.enable = 1
_C.HOOKS.EARLY_STOPPING.monitor = 'val_loss'
_C.HOOKS.EARLY_STOPPING.min_delta = 0.
_C.HOOKS.EARLY_STOPPING.patience = 5 
_C.HOOKS.EARLY_STOPPING.mode = 'min'
_C.HOOKS.EARLY_STOPPING.verbose = True

# ModelCheckpoint
_C.HOOKS.MODEL_CHECKPOINT = CN()
_C.HOOKS.MODEL_CHECKPOINT.enable = 1
_C.HOOKS.MODEL_CHECKPOINT.filepath = './output/checkpoints'
_C.HOOKS.MODEL_CHECKPOINT.monitor = 'val_loss'
_C.HOOKS.MODEL_CHECKPOINT.mode = 'min'


# ---------------------------------------------------------------------------- #
# Module template 
# ---------------------------------------------------------------------------- #

_C.MODULE_TEMPLATE = CN()
_C.MODULE_TEMPLATE.NAME = 'LightningTemplateModel'

# ---------------------------------------------------------------------------- #
# Trainer 
# ---------------------------------------------------------------------------- #

_C.TRAINER = CN()
_C.TRAINER.ACCUMULATE_GRAD_BATCHES = 1
_C.TRAINER.MIN_EPOCHS = 30
_C.TRAINER.MAX_EPOCHS = 1000
_C.TRAINER.GRAD_CLIP_VAL = 0.5 # clip gradient of which norm is larger than 0.5
_C.TRAINER.SHOW_PROGRESS_BAR = True # show progree bar
_C.TRAINER.ROW_LOG_INTERVAL = 100 # Every k batches lightning will make an entry in the metrics log
_C.TRAINER.LOG_SAVE_INTERVAL = 100 # Every k batches, lightning will write the new logs to disk, ie: save a .csv log file every 100 batches
_C.TRAINER.DEFAULT_SAVE_PATH = './output'
_C.TRAINER.LOG_GPU_MEMORY = "min_max" # 'min_max': log only the min/max utilization
_C.TRAINER.FAST_DEV_RUN = False # everything only with 1 training and 1 validation batch.
_C.TRAINER.LOGGER = CN()
_C.TRAINER.LOGGER.type = 'mlflow'
_C.TRAINER.LOGGER.experiment_name = 'default'
_C.TRAINER.LOGGER.tracking_uri = './output'

# _C.TRAINER.LOGGER = {
#     'type': 'test_tube',
#     'name': 'default',
#     'save_dir': './output',
#     'debug': False,
#     'create_git_tag': False,
# }    


# ---------------------------------------------------------------------------- #
# log
# ---------------------------------------------------------------------------- #

_C.LOG = CN()
_C.LOG.NAME = 'log.txt'

# ---------------------------------------------------------------------------- #
# Misc 
# ---------------------------------------------------------------------------- #

_C.OUTPUT_DIR = './output'
_C.SEED = 666
_C.CUDNN_BENCHMARK = False




#!/usr/bin/python
# -*- coding:utf-8 -*-

import logging
import time
import os


class Logger(object):
    '''
    封装后的logging
    '''

    def __init__(self, logger_name=None):
        '''
            指定保存日志的文件路径，日志级别，以及调用文件
            将日志存入到指定的文件中
        '''

        log_name = 'log.txt'
        # 创建一个logger
        self.logger = logging.getLogger(logger_name)
        self.logger.setLevel(logging.DEBUG)

        # 创建一个handler，用于写入日志文件
        self.log_time = time.strftime("%Y_%m_%d")
        self.log_dir = f'outputs_{self.log_time}'
        file_dir = os.path.join(os.getcwd(), self.log_dir)
        if not os.path.exists(file_dir):
            os.mkdir(file_dir)
        self.log_path = file_dir
        self.log_name = os.path.join(self.log_path, log_name)

        # fh = logging.FileHandler(self.log_name, 'a')  # 追加模式  这个是python2的
        fh = logging.FileHandler(self.log_name, 'a', encoding='utf-8')  # 这个是python3的
        fh.setLevel(logging.INFO)

        # 再创建一个handler，用于输出到控制台
        ch = logging.StreamHandler()
        ch.setLevel(logging.INFO)

        # 定义handler的输出格式
        formatter = logging.Formatter(
            '[%(asctime)s] %(filename)s -> %(funcName)s line:%(lineno)d [%(levelname)s] %(message)s')
        fh.setFormatter(formatter)
        ch.setFormatter(formatter)

        # 给logger添加handler
        self.logger.addHandler(fh)
        self.logger.addHandler(ch)

        #  添加下面一句，在记录日志之后移除句柄
        # self.logger.removeHandler(ch)
        # self.logger.removeHandler(fh)
        # 关闭打开的文件
        fh.close()
        ch.close()

    def getlog(self):
        return self.logger



"""
Runs a model on a single node across N-gpus.
"""
import sys
sys.path.append('.')
import argparse
import os
from argparse import ArgumentParser

import numpy as np
import torch
from pytorch_lightning import Trainer
from pytorch_lightning.callbacks import EarlyStopping, ModelCheckpoint
from pytorch_lightning.logging import TestTubeLogger, MLFlowLogger

from torchline.config import get_cfg
from torchline.engine import build_module_template
from torchline.utils import Logger

logger_print = Logger(__name__).getlog()
SEED = 2334
torch.manual_seed(SEED)
np.random.seed(SEED)


def setup(args):
    """
    Create configs and perform basic setups.
    """
    cfg = get_cfg()
    cfg.merge_from_file(args.config_file)
    cfg.merge_from_list(args.opts)
    cfg.freeze()
    
    if hasattr(args, "config_file"):
        logger_print.info(
            "Contents of args.config_file={}:\n{}".format(
                args.config_file, open(args.config_file, "r").read()
            )
        )
        
    # if not (hasattr(args, "eval_only") and args.eval_only):
    #     torch.backends.cudnn.benchmark = cfg.CUDNN_BENCHMARK
    
    logger_print.info("Running with full config:\n{}".format(cfg))
    return cfg

class MyTrainer(Trainer):
    def __init__(self, cfg, hparams):
        self.cfg = cfg
        self.hparams = hparams
        
        # hooks
        HOOKS = self.cfg.HOOKS
        params = {key: HOOKS.EARLY_STOPPING[key] for key in HOOKS.EARLY_STOPPING if key != 'enable'}
        early_stop_callback = EarlyStopping(**params) if HOOKS.EARLY_STOPPING['enable'] else False

        params = {key: HOOKS.MODEL_CHECKPOINT[key] for key in HOOKS.MODEL_CHECKPOINT if key != 'enable'}
        checkpoint_callback = ModelCheckpoint(**params)

        # logger
        LOGGER = self.cfg.TRAINER.LOGGER
        params = {key: LOGGER[key] for key in LOGGER if key != 'type'}
        if LOGGER['type'] == 'mlflow':
            logger = MLFlowLogger(**params)
        elif LOGGER['type'] == 'test_tube':
            logger = TestTubeLogger(**params)
        else:
            print(f"{LOGGER['type']} not supported")
            raise NotImplementedError

        # you can update trainer_params to change different parameters
        self.trainer_params = {
            'gpus': hparams.gpus,
            'use_amp': hparams.use_16bit,
            'distributed_backend': hparams.distributed_backend,

            # 'min_epochs' : cfg.TRAINER.MIN_EPOCHS,
            # 'max_epochs' : cfg.TRAINER.MAX_EPOCHS,
            'gradient_clip_val' : cfg.TRAINER.GRAD_CLIP_VAL,
            'show_progress_bar' : cfg.TRAINER.SHOW_PROGRESS_BAR,
            'row_log_interval' : cfg.TRAINER.ROW_LOG_INTERVAL,
            'log_save_interval' : cfg.TRAINER.LOG_SAVE_INTERVAL,
            'log_gpu_memory' : cfg.TRAINER.LOG_GPU_MEMORY,
            'default_save_path' : cfg.TRAINER.DEFAULT_SAVE_PATH,
            'fast_dev_run' : cfg.TRAINER.FAST_DEV_RUN,

            'logger': logger,
            'early_stop_callback': early_stop_callback,
            'checkpoint_callback': checkpoint_callback,

            'weights_summary': None
        }

        super(MyTrainer, self).__init__(
            **self.trainer_params
        )
    

def main(hparams):
    """
    Main training routine specific for this project
    :param hparams:
    """

    cfg = setup(hparams)

    # ------------------------
    # 1 INIT LIGHTNING MODEL
    # ------------------------
    model = build_module_template(cfg)

    # ------------------------
    # 2 INIT TRAINER
    # ------------------------
    trainer = MyTrainer(cfg, hparams)

    # ------------------------
    # 3 START TRAINING
    # ------------------------
    trainer.fit(model)


if __name__ == '__main__':
    # ------------------------
    # TRAINING ARGUMENTS
    # ------------------------
    # these are project-wide arguments

    root_dir = os.path.dirname(os.path.realpath(__file__))
    parent_parser = ArgumentParser(add_help=False)

    # gpu args
    parent_parser.add_argument(
        "--config_file", 
        default="", 
        metavar="FILE", 
        help="path to config file"
    )
    parent_parser.add_argument(
        '--gpus',
        type=int,
        default=2,
        help='how many gpus'
    )
    parent_parser.add_argument(
        '--distributed_backend',
        type=str,
        default='dp',
        help='supports three options dp, ddp, ddp2'
    )
    parent_parser.add_argument(
        '--use_16bit',
        dest='use_16bit',
        action='store_true',
        help='if true uses 16 bit precision'
    )
    parent_parser.add_argument(
        "opts",
        help="Modify config options using the command-line",
        default=None,
        nargs=argparse.REMAINDER,
    )

    # each LightningModule defines arguments relevant to it
    hyperparams = parent_parser.parse_args()

    # ---------------------
    # RUN TRAINING
    # ---------------------
    main(hyperparams)



from torchline.utils import Registry

DATASET_REGISTRY = Registry("DATASET")
DATASET_REGISTRY.__doc__ = """
Registry for dataset, i.e. torch.utils.data.Dataset.

The registered object will be called with `obj(cfg)`
"""


def build_data(cfg):
    """
    Built the dataset, defined by `cfg.DATASET.NAME`.
    """
    name = cfg.DATASET.NAME
    return DATASET_REGISTRY.get(name)(cfg)



#coding=utf-8

import torch
import torchvision
from torchvision.datasets import ImageFolder
from .transforms import build_transforms
from .build import DATASET_REGISTRY

__all__ = [
    'Skin10Dataset'
]

def load_clean_data(path, class_to_idx):
    imgs = []
    convert_index = {
    #'Acne_Vulgaris':
    1 : 0,
    #'Actinic_solar_Damage_Actinic_Keratosis':
    5 : 1,
    #'Atopic_Dermatitis_Eczema':
    16: 2,
    #'Basal_Cell_Carcinoma':
    17: 3,
    #'Compound_Nevus':
    22: 4,
    #'Onychomycosis':
    66: 5,
    #'Rosacea':
    75: 6,
    #'Seborrheic_Keratosis':
    81: 7,
    #'Stasis_Ulcer':
    85: 8,
    #'Tinea_Corporis':
    90: 9,
    }
    with open(path, 'r') as f:
        lines = f.readlines()
        for line in lines:
            class_name = line.rstrip().split('/')[-2]
            id_index = class_to_idx[class_name]
            id_index = convert_index[id_index]
            imgs.append((line.rstrip(), id_index))
    return imgs

@DATASET_REGISTRY.register()
class Skin10Dataset(ImageFolder):
    def __init__(self, cfg):
        super(Skin10Dataset, self).__init__(cfg.DATASET.DIR)
        self.cfg = cfg
        is_train = self.cfg.DATASET.IS_TRAIN
        if is_train:
            self.data_list = cfg.DATASET.TRAIN_LIST
        else:
            self.data_list = cfg.DATASET.TEST_LIST
        self.transforms = build_transforms(cfg)
        self.target_transforms = None
        self.samples = load_clean_data(self.data_list, self.class_to_idx)
        self.imgs = self.samples

    def __getitem__(self, index):
        path, target = self.samples[index]
        sample = self.loader(path)
        if self.transforms is not None:
            sample = self.transforms.transform(sample)
        if self.target_transforms is not None:
            target = self.target_transforms(target)

        return sample, target

    def __len__(self):
        return len(self.samples)



import torch
from torchline.utils import Registry, Logger

SAMPLER_REGISTRY = Registry('SAMPLER')
SAMPLER_REGISTRY.__doc__ = """
Registry for dataset sampler, i.e. torch.utils.data.Sampler.

The registered object will be called with `obj(cfg)`
"""


def build_sampler(cfg):
    """
    Built the dataset sampler, defined by `cfg.DATASET.NAME`.
    """
    is_train = cfg.DATASET.IS_TRAIN
    if is_train:
        name = cfg.DATALOADER.SAMPLER_TRAIN
    else:
        name = cfg.DATALOADER.SAMPLER_TEST
    if name == 'default':
        return None
    return SAMPLER_REGISTRY.get(name)(cfg)

@SAMPLER_REGISTRY.register()
class ImbalancedDatasetSampler(torch.utils.data.sampler.Sampler):
    """Samples elements randomly from a given list of indices for imbalanced dataset
    Arguments:
        indices (list, optional): a list of indices
        num_samples (int, optional): number of samples to draw
    """

    def __init__(self, dataset, indices=None, num_samples=None):
                
        # if indices is not provided, 
        # all elements in the dataset will be considered
        self.indices = list(range(len(dataset))) \
            if indices is None else indices
            
        # if num_samples is not provided, 
        # draw `len(indices)` samples in each iteration
        self.num_samples = len(self.indices) \
            if num_samples is None else num_samples
            
        # distribution of classes in the dataset 
        label_to_count = {}
        for idx in self.indices:
            label = self._get_label(dataset, idx)
            if label in label_to_count:
                label_to_count[label] += 1
            else:
                label_to_count[label] = 1
                
        # weight for each sample
        weights = [1.0 / label_to_count[self._get_label(dataset, idx)]
                   for idx in self.indices]
        self.weights = torch.DoubleTensor(weights)

    def _get_label(self, dataset, idx):
        dataset_type = type(dataset)
        if dataset_type is torchvision.datasets.MNIST:
            return dataset.train_labels[idx].item()
        elif dataset_type is torchvision.datasets.ImageFolder:
            return dataset.imgs[idx][1]
        else:
            try:
                return dataset.imgs[idx][-1]
            except Exception as e:
                print(str(e))
                raise NotImplementedError
                
    def __iter__(self):
        return (self.indices[i] for i in torch.multinomial(
            self.weights, self.num_samples, replacement=True))

    def __len__(self):
        return self.num_samples


@SAMPLER_REGISTRY.register()
class DiffWeightedRandomSampler(torch.utils.data.sampler.Sampler):
    r"""
    Samples elements from a given list of indices with given probabilities (weights), with replacement.

    Arguments:
        weights (sequence) : a sequence of weights, not necessary summing up to one
        num_samples (int): number of samples to draw

    """

    def __init__(self, indices, weights, num_samples=0):
        if not isinstance(num_samples, int) or isinstance(num_samples, bool):
            raise ValueError("num_samples should be a non-negative integeral "
                             "value, but got num_samples={}".format(num_samples))
        self.indices = indices
        weights = [ weights[i] for i in self.indices ]
        self.weights = torch.DoubleTensor(weights)
        if num_samples == 0:
            self.num_samples = len(self.weights)
        else:
            self.num_samples = num_samples
        self.replacement = True

    def __iter__(self):
        return (self.indices[i] for i in torch.multinomial(self.weights, self.num_samples, self.replacement))

    def __len__(self):
        return self.num_samples


#coding=utf-8

import torch
import torchvision
from torchvision.datasets import ImageFolder

from .build import DATASET_REGISTRY
from .transforms import build_transforms

__all__ = [
    'Skin100Dataset'
]

def load_clean_data(path, class_to_idx):
    imgs = []
    with open(path, 'r') as f:
        lines = f.readlines()
        for line in lines:
            class_name = line.rstrip().split('/')[-2]
            id_index = class_to_idx[class_name]
            imgs.append((line.rstrip(), id_index))
    return imgs

@DATASET_REGISTRY.register()
class Skin100Dataset(ImageFolder):
    def __init__(self, args, is_train):
        super(Skin100Dataset, self).__init__(args.data_root)
        self.args = args
        if is_train:
            self.data_list = args.trainlist
        else:
            self.data_list = args.testlist
        self.transform = get_transform_ops(args, is_train)
        self.target_transform = None
        self.samples = load_clean_data(self.data_list, self.class_to_idx)
        self.imgs = self.samples

    def __getitem__(self, index):
        path, target = self.samples[index]
        sample = self.loader(path)
        if self.transform is not None:
            sample = self.transform(sample)
        if self.target_transform is not None:
            target = self.target_transform(target)

        return sample, target

    def __len__(self):
        return len(self.samples)



import torch
import torch.nn as nn
from torchvision import models

from .build import META_ARCH_REGISTRY

__all__ = [
    'Resnet18',
    'Resnet34',
    'Resnet50',
    'Resnet101',
    'Resnet152',
]

class Identity(nn.Module):
    def __init__(self):
        super(Identity, self).__init__()

    def forward(self, x):
        return x

@META_ARCH_REGISTRY.register()
class Resnet50(nn.Module):
    def __init__(self, cfg):
        super(Resnet50, self).__init__()
        self.cfg = cfg
        self.num_classes = cfg.MODEL.CLASSES
        name = cfg.MODEL.META_ARCH
        pretrained = cfg.MODEL.PRETRAINED
        model = eval(f"models.{name.lower()}(pretrained={pretrained})")
        model.fc = Identity()
        model = list(model.children())
        self.stem = nn.Sequential(*model[:4])
        self.layer1 = nn.Sequential(*model[4])
        self.layer2 = nn.Sequential(*model[5])
        self.layer3 = nn.Sequential(*model[6])
        self.layer4 = nn.Sequential(*model[7])

        self.avg_pool = nn.AdaptiveAvgPool2d((1,1))
        self.clf = nn.Linear(in_features=2048, out_features=self.num_classes)
    
    def extract_features(self, x):
        assert len(self.cfg.MODEL.FEATURES) >= 1
        features = {}
        features['stem'] = self.stem(x)
        features['f1'] = self.layer1(features['stem'])
        features['f2'] = self.layer2(features['f1'])
        features['f3'] = self.layer3(features['f2'])
        features['f4'] = self.layer4(features['f3'])
        if len(self.cfg.MODEL.FEATURES) == 1:
            final_feature = features[self.cfg.MODEL.FEATURES[0]]
        else:
            final_feature = features[self.cfg.MODEL.FEATURES[0]]
            for f in self.cfg.MODEL.FEATURES[1:]:
                final_feature += features[f]
        return final_feature

    
    def forward(self, x):
        '''
        params: 
            x (tensor): N*c*h*w

        return:
            predictions (tensor): N*classes
        '''
        bs= x.shape[0]
        features = self.extract_features(x)
        predictions = self.avg_pool(features).view(bs, -1)
        predictions = self.clf(predictions).view(bs, -1)

        return predictions



@META_ARCH_REGISTRY.register()
class Resnet18(Resnet50):
    def __init__(self, cfg):
        super(Resnet18, self).__init__(cfg)


@META_ARCH_REGISTRY.register()
class Resnet34(Resnet50):
    def __init__(self, cfg):
        super(Resnet34, self).__init__(cfg)


@META_ARCH_REGISTRY.register()
class Resnet101(Resnet50):
    def __init__(self, cfg):
        super(Resnet101, self).__init__(cfg)


@META_ARCH_REGISTRY.register()
class Resnet152(Resnet50):
    def __init__(self, cfg):
        super(Resnet152, self).__init__(cfg)




#coding=utf-8
import torch
import numpy as np
import math
import random
import torchvision
from torchvision import transforms
from . import autoaugment
from .data_utils import Cutout, RandomErasing
from torchline.utils import Registry, Logger

TRANSFORMS_REGISTRY = Registry('TRANSFORMS')
TRANSFORMS_REGISTRY.__doc__ = """
Registry for data transform functions, i.e. torchvision.transforms

The registered object will be called with `obj(cfg)`
"""

LABEL_TRANSFORMS_REGISTRY = Registry('LABEL_TRANSFORMS')
LABEL_TRANSFORMS_REGISTRY.__doc__ = """
Registry for label transform functions, i.e. torchvision.transforms

The registered object will be called with `obj(cfg)`
"""

__all__ = [
    'build_transforms',
    'build_label_transforms',
    'TRANSFORMS_REGISTRY',
    'LABEL_TRANSFORMS_REGISTRY',
    'DefaultTransforms'
]

logger_print = Logger(__name__).getlog()

def build_transforms(cfg):
    """
    Built the transforms, defined by `cfg.TRANSFORMS.NAME`.
    """
    name = cfg.TRANSFORMS.NAME
    return TRANSFORMS_REGISTRY.get(name)(cfg)
    
def build_label_transforms(cfg):
    """
    Built the label transforms, defined by `cfg.LABEL_TRANSFORMS.NAME`.
    """
    name = cfg.LABEL_TRANSFORMS.NAME
    if name == 'default':
        return None
    return LABEL_TRANSFORMS_REGISTRY.get(name)(cfg)

@TRANSFORMS_REGISTRY.register()
class DefaultTransforms:
    def __init__(self, cfg):
        self.cfg = cfg
        self.is_train = cfg.DATASET.IS_TRAIN
        self.mean = cfg.TRANSFORMS.TENSOR.NORMALIZATION.mean
        self.std = cfg.TRANSFORMS.TENSOR.NORMALIZATION.std
        self.img_size = cfg.INPUT.SIZE
        self.min_edge_size = min(self.img_size)
        self.normalize = transforms.Normalize(self.mean, self.std)
        self.transform = self.get_transform()

    def get_transform(self):
        # validation transform
        if not self.is_train: 
            logger_print.info('Generating validation transform ...')
            transform = transforms.Compose([
                transforms.Resize(self.img_size),
                transforms.ToTensor(),
                self.normalize
            ])
            return transform

        logger_print.info('Generating training transform ...')
        if self.cfg.TRANSFORMS.IMG.AUG_IMAGENET:
            logger_print.info('Using imagenet augmentation')
            transform = transforms.Compose([
                transforms.Resize(self.min_edge_size+1),
                transforms.RandomCrop(self.min_edge_size),
                autoaugment.ImageNetPolicy(),
                transforms.ToTensor(),
                self.normalize
            ])
        elif self.cfg.TRANSFORMS.IMG.AUG_CIFAR:
            logger_print.info('Using cifar augmentation')
            transform = transforms.Compose([
                transforms.Resize(self.min_edge_size+1),
                transforms.RandomCrop(self.min_edge_size),
                autoaugment.CIFAR10Policy(),
                transforms.ToTensor(),
                self.normalize
            ])
        else:
            transform = self.read_transform_from_cfg()
        return transform

    def read_transform_from_cfg(self):
        transform_list = []
        self.check_resize_options()
        img_transforms = self.cfg.TRANSFORMS.IMG

        # resize and crop opertaion
        if img_transforms.RANDOM_RESIZED_CROP.enable:
            transform_list.append(transforms.RandomResizedCrop(self.img_size))
        elif img_transforms.RESIZE.enable:
            transform_list.append(transforms.Resize(self.min_edge_size))
            if img_transforms.RANDOM_CROP.enable:
                transform_list.append(transforms.RandomCrop(self.min_edge_size))
            elif img_transforms.CENTER_CROP.enable:
                transform_list.append(transforms.CenterCrop(self.min_edge_size))

        # ColorJitter
        if img_transforms.COLOR_JITTER.enable:
            params = {key: img_transforms.COLOR_JITTER[key] for key in img_transforms.COLOR_JITTER 
                            if key != 'enable'}
            transform_list.append(transforms.ColorJitter(**params))

        # horizontal flip
        if img_transforms.RANDOM_HORIZONTAL_FLIP.enable:
            p = img_transforms.RANDOM_HORIZONTAL_FLIP.p
            transform_list.append(transforms.RandomHorizontalFlip(p))
        
        # vertical flip
        if img_transforms.RANDOM_VERTICAL_FLIP.enable:
            p = img_transforms.RANDOM_VERTICAL_FLIP.p
            transform_list.append(transforms.RandomVerticalFlip(p))

        # rotation
        if img_transforms.RANDOM_ROTATION.enable:
            degrees = img_transforms.RANDOM_ROTATION.degrees
            transform_list.append(transforms.RandomRotation(degrees))
        return transforms.Compose(transform_list)

    def check_conflict_options(self):
        count = self.cfg.TRANSFORMS.IMG.RANDOM_RESIZED_CROP.enable + \
                self.cfg.TRANSFORMS.IMG.RESIZE.enable
        assert count <= 1, 'You can only use one resize transform operation'

        count = self.cfg.TRANSFORMS.IMG.RANDOM_CROP.enable + \
                self.cfg.TRANSFORMS.IMG.CENTER_CROP.enable
        assert count <= 1, 'You can only use one crop transform operation'
